@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 """Exosite RPC API Command Line Interface
    Provides command line access to the Remote Procedure Call API:
-   http://developers.exosite.com/display/OP/Remote+Procedure+Call+API
+   https://github.com/exosite/api/tree/master/rpc
 
 Usage:
   exo [--help] [options] <command> [<args> ...]
@@ -26,6 +26,7 @@ Commands:
   spark
   copy
   diff
+  data
 
 Options:
   --host=<host>        OneP URL. Default is $EXO_HOST or m2.exosite.com
@@ -33,6 +34,8 @@ Options:
   --httptimeout=<sec>  HTTP timeout [default: 60]
   --https              Enable HTTPS
   --debug              Show info like stack traces
+  --debughttp          Turn on debug level loggin in pyonep
+  --discreet           Obfuscate RIDs in stdout and stderr
   -h --help            Show this screen
   -v --version         Show version
 
@@ -53,17 +56,19 @@ from pprint import pprint
 from operator import itemgetter
 import logging
 from collections import defaultdict
+import StringIO
 # python 2.6 support
 try:
     from collections import OrderedDict
 except ImportError:
     from ordereddict import OrderedDict
-
 import itertools
 import math
 
 from docopt import docopt
 from dateutil import parser
+import requests
+
 from pyonep import onep
 import pyonep
 try:
@@ -71,9 +76,6 @@ try:
 except:
     from exoline import __version__
 
-logging.basicConfig(stream=sys.stderr)
-logging.getLogger("pyonep.onep").setLevel(logging.ERROR)
-
 DEFAULT_HOST = 'm2.exosite.com'
 DEFAULT_PORT = '80'
 DEFAULT_PORT_HTTPS = '443'
@@ -88,6 +90,7 @@ Options:
     --limit=<limit>          number of data points to read [default: 1]
     --start=<time>
     --end=<time>             start and end times (see details below)
+    --sort=<order>           asc or desc [default: desc]
     --selection=all|autowindow|givenwindow  downsample method [default: all]
     --format=csv|raw         output format [default: csv]
     --timeformat=unix|human|iso8601
@@ -155,7 +158,8 @@ Options:
                    all available keys are returned.''',
     'update':
         '''Update a resource from a json description passed on stdin.\n\nUsage:
-    exo [options] update <cik> [<rid>] -
+    exo [options] update <cik> (<rid> - | -)
+
 
 Restrictions:
 
@@ -203,9 +207,8 @@ Options:
     exo [options] usage <cik> [<rid>] --start=<time> [--end=<time>]
 
     {{ startend }}''',
-    'tree':
-        '''Display a resource's descendants.\n\nUsage:
-    exo [options] tree [--verbose] [--hide-keys] <cik>
+    'tree': '''Display a resource's descendants.\n\nUsage:
+    exo [options] tree [--verbose] <cik>
 
     --counts       show item counts (from info.storage.counts)
     --level=<num>  depth to traverse, omit or -1 for no limit [default: -1]''',
@@ -230,7 +233,8 @@ Options:
     Returns CIK of the copy. NOTE: copy excludes all data in dataports.
 
 Options:
-    --cikonly  show unlabeled CIK by itself''',
+    --cikonly  show unlabeled CIK by itself
+    {{ helpoption }}''',
     'diff': '''Show differences between two clients.\n\nUsage:
     exo [options] diff <cik> <cik2>
 
@@ -240,7 +244,18 @@ Options:
 
 Options:
     --full         compare all info, even usage, data counts, etc.
-    --no-children  don't compare children'''
+    --no-children  don't compare children
+    {{ helpoption }}''',
+    'ip': '''Get IP address of the server.\n\nUsage:
+    exo [options] ip''',
+    'data': '''Read or write with the HTTP Data API.\n\nUsage:
+    exo [options] data <cik> [--write=<alias,value> ...] [--read=<alias> ...]
+
+    Writes and/or reads a set of dataports.
+    If only --write arguments are specified, the call is a write.
+    If only --read arguments are specified, the call is a read.
+    If both --write and --read arguments are specified, the hybrid
+        write/read API is used. Writes are executed before reads.'''
 }
 
 # shared sections of documentation
@@ -281,6 +296,7 @@ class ExolineOnepV1(onep.OnepV1):
 class ExoRPC():
     '''Wrapper for pyonep RPC API.
     Raises exceptions on error and provides some reasonable defaults.'''
+    regex_rid = re.compile("[0-9a-zA-Z]{40}")
     def __init__(self,
             host=DEFAULT_HOST,
             port=None,
@@ -345,6 +361,35 @@ class ExoRPC():
             options['endtime'] = int(endtime)
         return options
 
+    def lookup_shortcut(self, cik):
+        '''Look up what was passed for cik in config file
+            if it doesn't look like a CIK.'''
+        if self.regex_rid.match(cik) is None:
+            # if cik doesn't look like a cik, maybe it's a shortcut
+            configfile = os.path.join(os.environ['HOME'], '.exoline')
+            try:
+                import yaml
+                with open(configfile) as f:
+                    config = yaml.safe_load(f)
+                    if 'keys' in config:
+                        if cik in config['keys']:
+                            return config['keys'][cik].strip()
+                        else:
+                            raise ExoException('No CIK shortcut {0}\n{1}'.format(
+                                cik,
+                                '\n'.join(config['keys'])))
+                    else:
+                        raise ExoException('Tried a CIK shortcut {0}, but found no keys in {1}'.format(
+                            cik,
+                            configfile))
+            except IOError as ex:
+                raise ExoException(
+                    'Tried a CIK shortcut {0}, but couldn\'t open {1}'.format(
+                    cik,
+                    configfile))
+        else:
+            return cik
+
     def read(self,
              cik,
              rid,
@@ -361,7 +406,7 @@ class ExoRPC():
         self._raise_for_response(isok, response)
         return response
 
-    def _combinereads(self, reads):
+    def _combinereads(self, reads, sort):
         '''
         >>> exo = ExoRPC()
         >>> exo._combinereads([[[2, 'a'], [1, 'b']]])
@@ -407,7 +452,12 @@ class ExoRPC():
                     if v is not None:
                         curi[i] -= 1
 
-            combined.sort(key=itemgetter(0), reverse=True)
+            if sort == 'desc':
+                reverse = True
+            else:
+                reverse = False
+
+            combined.sort(key=itemgetter(0), reverse=reverse)
             return combined
 
     def readmult(self,
@@ -428,7 +478,7 @@ class ExoRPC():
 
         def _read(cik, rids, options):
             responses = self._exomult(cik, [['read', rid, options] for rid in rids])
-            return self._combinereads(responses)
+            return self._combinereads(responses, options['sort'])
 
         def _read_chunk(cik, rids, options, start, end):
             chunkoptions = options.copy()
@@ -483,7 +533,7 @@ class ExoRPC():
                     yield r
 
     def write(self, cik, rid, value):
-        isok, response = self.exo.write(cik, rid, value, {})
+        isok, response = self.exo.write(cik, rid, value)
         self._raise_for_response(isok, response)
 
     def record(self, cik, rid, entries):
@@ -498,7 +548,7 @@ class ExoRPC():
         return response
 
     def update(self, cik, rid, desc):
-        isok, response = self.exo.create(cik, rid, desc)
+        isok, response = self.exo.update(cik, rid, desc)
         self._raise_for_response(isok, response)
         return response
 
@@ -631,20 +681,12 @@ class ExoRPC():
             print("{0}:{1} {2}".format(
                   metrics[i], ' ' * (maxlen - len(metrics[i])), r))
 
-    def _disp_key(self, cli_args, k):
-        if cli_args['--hide-keys']:
-            # number of digits to show
-            num = 20
-            return k[:num] + '0' * len(k[num:])
-        else:
-            return k
-
     def _print_node(self, rid, info, aliases, cli_args, spacer, islast):
         typ = info['basic']['type']
         if typ == 'client':
-            id = 'cik: ' + self._disp_key(cli_args, info['key'])
+            id = 'cik: ' + info['key']
         else:
-            id = 'rid: ' + self._disp_key(cli_args, rid)
+            id = 'rid: ' + rid
         name = info['description']['name']
         try:
             # Units are a portals only thing
@@ -693,7 +735,7 @@ class ExoRPC():
                 ridopt = '--verbose'
             else:
                 ridopt = True
-        add_opt(ridopt, 'rid', self._disp_key(cli_args, rid))
+        add_opt(ridopt, 'rid', rid)
         add_opt('--verbose', 'unit', units)
         if 'storage' in info and 'count' in info['storage']:
             add_opt(True, 'count', info['storage']['count'])
@@ -940,6 +982,7 @@ class ExoRPC():
 
         # read in the whole client to copy at once
         if infotree is None:
+            destcik = self.lookup_shortcut(destcik)
             infotree = self._infotree(cik, options={})
 
         # check counts
@@ -1074,6 +1117,8 @@ class ExoRPC():
     def diff(self, cik1, cik2, full=False, nochildren=False):
         '''Show differences between two ciks.'''
 
+        cik2 = self.lookup_shortcut(cik2)
+
         # list of info "keypaths" to not include in comparison
         # only the last item in the list is removed. E.g. for a
         # keypath of ['counts', 'disk'], only the 'disk' key is
@@ -1155,6 +1200,52 @@ class ExoRPC():
         return options
 
 
+class ExoData():
+    '''Implements the Data Interface API
+    https://github.com/exosite/api/tree/master/data'''
+
+    def __init__(self, url='http://m2.exosite.com'):
+        self.url = url
+
+    def raise_for_status(self, r):
+        try:
+            r.raise_for_status()
+        except Exception as ex:
+            raise ExoException(str(ex))
+
+    def read(self, cik, aliases):
+        headers = {'X-Exosite-CIK': cik,
+                   'Accept': 'application/x-www-form-urlencoded; charset=utf-8'}
+        url = self.url + '/onep:v1/stack/alias?' + '&'.join(aliases)
+        r = requests.get(url, headers=headers)
+        self.raise_for_status(r)
+        return r.text
+
+    def write(self, cik, alias_values):
+        headers = {'X-Exosite-CIK': cik,
+                   'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8'}
+        url = self.url + '/onep:v1/stack/alias'
+        data = '&'.join(['='.join(t) for t in alias_values])
+        r = requests.post(url, headers=headers, data=data)
+        self.raise_for_status(r)
+        return r.text
+
+    def writeread(self, cik, alias_values, aliases):
+        headers = {'X-Exosite-CIK': cik,
+                   'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
+                   'Accept': 'application/x-www-form-urlencoded; charset=utf-8'}
+        url = self.url + '/onep:v1/stack/alias?' + '&'.join(aliases)
+        data = '&'.join(['='.join(t) for t in alias_values])
+        r = requests.post(url, headers=headers, data=data)
+        self.raise_for_status(r)
+        return r.text
+
+    def ip(self):
+        r = requests.get(self.url + '/ip')
+        r.raise_for_status()
+        return r.text
+
+
 def parse_ts(s):
     return int(time.mktime(parser.parse(s).timetuple()))
 
@@ -1378,11 +1469,13 @@ def read_cmd(er, cik, rids, args):
         while True:
             results = er.readmult(cik,
                                   rids,
+                                  sort='desc',
                                   # Read all points that arrived since last
                                   # read time. Could also be now - last_t?
                                   limit=sleep_seconds * 3,
                                   starttime=last_t + 1)
             results = list(results)
+            results.reverse()
 
             for t, v in results:
                 printline(t, v)
@@ -1398,7 +1491,7 @@ def read_cmd(er, cik, rids, args):
                 "--chunkhours requires --start and --end be set")
         for t, v in er.readmult(cik,
                                 rids,
-                                sort='desc',
+                                sort=args['--sort'],
                                 starttime=start,
                                 endtime=end,
                                 limit=limit,
@@ -1414,50 +1507,30 @@ def pretty_print(arg):
     print(json.dumps(arg, sort_keys=True, indent=4, separators=(',', ': ')))
 
 
+
+
 def handle_args(cmd, args):
     er = ExoRPC(host=args['--host'], port=args['--port'], https=args['--https'], httptimeout=args["--httptimeout"])
+    if cmd in ['ip', 'data']:
+        if args['--https'] is True or args['--port'] is not None or args['--debughttp'] is True:
+            # TODO: support these
+            raise ExoException('--https, --port, and --debughttp are not supported for ip and data commands.')
+        ed = ExoData(url='http://' + args['--host'])
 
-    regex_rid = re.compile("[0-9a-zA-Z]{40}")
-    cik = args['<cik>']
-
-    def cik_or_shortcut(cik):
-        '''Look up what was passed for <cik> in config
-           if it doesn't look like a CIK.'''
-        if regex_rid.match(cik) is None:
-            # if cik doesn't look like a cik, maybe it's a shortcut
-            configfile = os.path.join(os.environ['HOME'], '.exoline')
-            try:
-                import yaml
-                with open(configfile) as f:
-                    config = yaml.safe_load(f)
-                    if 'keys' in config:
-                        if cik in config['keys']:
-                            return config['keys'][cik].strip()
-                        else:
-                            raise ExoException('No CIK shortcut {0}\n{1}'.format(
-                                cik,
-                                '\n'.join(config['keys'])))
-                    else:
-                        raise ExoException('Tried a CIK shortcut {0}, but found no keys in {1}'.format(
-                            cik,
-                            configfile))
-            except IOError as ex:
-                raise ExoException(
-                    'Tried a CIK shortcut {0}, but couldn\'t open {1}'.format(
-                    cik,
-                    configfile))
+    if '<cik>' in args:
+        cik = args['<cik>']
+        if type(cik) is list:
+            cik = [er.lookup_shortcut(c) for c in cik]
         else:
-            return cik
-
-    if type(cik) is list:
-        cik = [cik_or_shortcut(c) for c in cik]
+            cik = er.lookup_shortcut(cik)
     else:
-        cik = cik_or_shortcut(cik)
+        # for data ip command
+        cik = None
 
     def rid_or_alias(rid):
         '''Translate what was passed for <rid> to an alias object if
            it doesn't look like a RID.'''
-        if regex_rid.match(rid) is None:
+        if er.regex_rid.match(rid) is None:
             return {"alias": rid}
         else:
             return rid
@@ -1487,27 +1560,32 @@ def handle_args(cmd, args):
         if interval is None:
             entries = []
             # split timestamp, value
+            has_errors = False
             if args['-']:
-                tvalues = sys.stdin.readlines()
+                headers = ['timestamp', 'value']
+                dr = csv.DictReader(sys.stdin, headers)
+                for row in dr:
+                    # TODO: handle something other than unix timestamps
+                    entries.append([int(row['timestamp']), row['value']])
             else:
                 tvalues = args['--value']
+                reentry = re.compile('(-?\d+),(.*)')
+                for tv in tvalues:
+                    match = reentry.match(tv)
+                    if match is None:
+                        try:
+                            t, v = tv.split(',')
+                            t = parse_ts(t)
+                            entries.append([t, v])
+                        except Exception:
+                            sys.stderr.write(
+                                'Line not in <timestamp>,<value> format: {0}'.format(tv))
+                            has_errors = True
+                    else:
+                        g = match.groups()
+                        # TODO: handle something other than unix timestamps
+                        entries.append([int(g[0]), g[1]])
 
-            reentry = re.compile('(-?\d+),(.*)')
-            has_errors = False
-            for tv in tvalues:
-                match = reentry.match(tv)
-                if match is None:
-                    try:
-                        t, v = tv.split(',')
-                        t = parse_ts(t)
-                        entries.append([t, v])
-                    except Exception:
-                        sys.stderr.write(
-                            'Line not in <timestamp>,<value> format: {0}'.format(tv))
-                        has_errors = True
-                else:
-                    g = match.groups()
-                    entries.append([int(g[0]), g[1]])
             if has_errors or len(entries) == 0:
                 raise ExoException("Problems with input.")
             else:
@@ -1559,7 +1637,6 @@ def handle_args(cmd, args):
             er.map(cik, rid, args['--alias'])
             if not ridonly:
                 print("alias: {0}".format(args['--alias']))
-
     elif cmd == 'update':
         s = sys.stdin.read()
         try:
@@ -1573,8 +1650,10 @@ def handle_args(cmd, args):
         er.unmap(cik, args['<alias>'])
     elif cmd == 'lookup':
         # look up by cik or alias
-        if args['--cik'] is not None:
-            rid = er.lookup_rid(cik, args['--cik'])
+        cik_to_find = args['--cik']
+        if cik_to_find is not None:
+            cik_to_find = er.lookup_shortcut(cik_to_find)
+            rid = er.lookup_rid(cik, cik_to_find)
             if rid is not None:
                 pr(rid)
         else:
@@ -1664,10 +1743,47 @@ def handle_args(cmd, args):
                         nochildren=args['--no-children'])
         if diffs is not None:
             print(diffs)
+    elif cmd == 'ip':
+        pr(ed.ip())
+    elif cmd == 'data':
+        reads = args['--read']
+        writes = args['--write']
+        def get_alias_values(writes):
+            # TODO: support values with commas
+            alias_values = []
+            re_assign = re.compile('(.*),(.*)')
+            for w in writes:
+                if w.count(',') > 1:
+                    raise ExoException('Values with commas not supported yet.')
+                m = re_assign.match(w)
+                if m is None or len(m.groups()) != 2:
+                    raise ExoException("Bad alias assignment format")
+                alias_values.append(m.groups())
+            return alias_values
+
+        if len(reads) > 0 and len(writes) > 0:
+            alias_values = get_alias_values(writes)
+            print(ed.writeread(cik, alias_values, reads))
+        elif len(reads) > 0:
+            print(ed.read(cik, reads))
+        elif len(writes) > 0:
+            alias_values = get_alias_values(writes)
+            ed.write(cik, alias_values)
     else:
         raise ExoException("Command not handled")
 
 
+class DiscreetFilter(object):
+    '''Filter stdin/stdout to hide anything that looks like
+       an RID'''
+    def __init__(self, out):
+        self.out = out
+        self.ridre = re.compile('([a-fA-F0-9]{20})([a-fA-F0-9]{20})')
+
+    def write(self, message):
+        self.out.write(self.ridre.sub(r'\g<1>01234567890123456789',
+                                      message))
+
 def cmd(argv=None, stdin=None, stdout=None, stderr=None):
     '''Wrap the command line interface. Globally redirects args
     and io so that the application can be tested externally.'''
@@ -1695,10 +1811,21 @@ def cmd(argv=None, stdin=None, stdout=None, stderr=None):
     else:
         print('Unknown command {0}. Try "exo --help"'.format(cmd))
         return 1
-
     # merge command-specific arguments into general arguments
     args.update(args_cmd)
 
+    # turn on stdout/stderr filtering
+    if args['--discreet']:
+        sys.stdout = DiscreetFilter(sys.stdout)
+        sys.stderr = DiscreetFilter(sys.stderr)
+
+    # configure logging
+    logging.basicConfig(stream=sys.stderr)
+    logging.getLogger("pyonep.onep").setLevel(logging.ERROR)
+    if args['--debughttp']:
+        # TODO: log debug level messages to stdout
+        logging.getLogger("pyonep.onep").setLevel(logging.DEBUG)
+
     # substitute environment variables
     if args['--host'] is None:
         args['--host'] = os.environ.get('EXO_HOST', DEFAULT_HOST)
@@ -1731,5 +1858,33 @@ def cmd(argv=None, stdin=None, stdout=None, stderr=None):
     return 0
 
 
+class CmdResult():
+    def __init__(self, exitcode, stdout, stderr):
+        self.exitcode = exitcode
+        self.stdout = stdout
+        self.stderr = stderr
+
+
+def run(argv, stdin):
+    '''Runs an exoline command, translating stdin from
+    string and stdout to string. Returns a CmdResult.'''
+    if type(stdin) is str:
+        sio = StringIO.StringIO()
+        sio.write(stdin)
+        sio.seek(0)
+        stdin = sio
+    stdout = StringIO.StringIO()
+    stderr = StringIO.StringIO()
+
+    # unicode causes problems in docopt
+    argv = [str(a) for a in argv]
+    exitcode = cmd(argv=argv, stdin=stdin, stdout=stdout, stderr=stderr)
+    stdout.seek(0)
+    stdout = stdout.read().strip()  # strip to get rid of leading newline
+    stderr.seek(0)
+    stderr = stderr.read().strip()
+    return CmdResult(exitcode, stdout, stderr)
+
+
 if __name__ == '__main__':
     sys.exit(cmd(sys.argv))
