@@ -9,65 +9,40 @@ import json
 import re
 import time
 from datetime import datetime
-import StringIO
 import logging
 from unittest import TestCase
 
 from ..exoline import exo
 from ..exoline.exo import ExolineOnepV1
 
-
 try:
     from testconfig import config
 except:
     sys.stderr.write(
         "Copy testconfig.py.template to testconfig.py and set portalcik.")
 
-class CmdResult():
-    def __init__(self, exitcode, stdout, stderr):
-        self.exitcode = exitcode
-        self.stdout = stdout
-        self.stderr = stderr
 
 logging.basicConfig(stream=sys.stderr)
 logging.getLogger("TestRPC").setLevel(logging.DEBUG)
-logging.getLogger("_cmd").setLevel(logging.DEBUG)
+logging.getLogger("run").setLevel(logging.DEBUG)
 logging.getLogger("pyonep.onep").setLevel(logging.ERROR)
-log = logging.getLogger("_cmd")
+log = logging.getLogger("run")
+
 
 def abbrev(s, length=1000):
     if len(s) > length:
-        s = s[:length/2] + '\n...\n' + s[-length/2:]
+        s = s[:length / 2] + '\n...\n' + s[-length / 2:]
     return s
 
-def _cmd(argv, stdin, block=True):
-    '''Runs an exoline command, translating stdin from
-    string and stdout to string. Returns a CmdResult.'''
-    if True:
-        log.debug(' '.join([str(a) for a in argv]))
-        if stdin is not None:
-            log.debug('    stdin: ' + abbrev(stdin))
-    if type(stdin) is str:
-        sio = StringIO.StringIO()
-        sio.write(stdin)
-        sio.seek(0)
-        stdin = sio
-    stdout = StringIO.StringIO()
-    stderr = StringIO.StringIO()
-
-    # unicode causes problems in docopt
-    argv = [str(a) for a in argv]
-    exitcode = exo.cmd(argv=argv, stdin=stdin, stdout=stdout, stderr=stderr)
-    stdout.seek(0)
-    stdout = stdout.read().strip()  # strip to get rid of leading newline
-    stderr.seek(0)
-    stderr = stderr.read().strip()
-    return CmdResult(exitcode, stdout, stderr)
-
 
 def rpc(*args, **kwargs):
     stdin = kwargs.get('stdin', None)
-    return _cmd(['exo'] + list(args), stdin=stdin)
+    if True:
+        argv = ['exo'] + list(args)
+        log.debug(' '.join([str(a) for a in argv]))
+        if stdin is not None:
+            log.debug('    stdin: ' + abbrev(stdin))
+    return exo.run(argv, stdin=stdin)
 
 
 class Resource():
@@ -103,16 +78,19 @@ class TestRPC(TestCase):
     RE_RID = '[0-9a-f]{40}'
 
     def _logall(self, r):
-        self.l('stdout: {0}\nstderr: {1}'.format(abbrev(r.stdout), abbrev(r.stderr)))
+        self.l('stdout: {0}\nstdout length:{1}\nstderr: {2}\nstderr length:{3}'.format(abbrev(r.stdout),
+                                                                                       len(r.stdout),
+                                                                                       abbrev(r.stderr),
+                                                                                       len(r.stderr)))
 
     def _stdre(self, r, msg, search, match, stderr=False):
         std, label = (r.stderr, "stderr") if stderr else (r.stdout, "stdout")
         if search is not None:
             self.assertTrue(re.search(search, std, flags=re.MULTILINE) is not None,
-                msg + ' - failed to find {0}\n'.format(search) + label + ':' + std + '\nsearch expression: ' + search)
+                            msg + u' - failed to find in {1}:\n{2}\nsearch expression:\n{0}\nlengths: {3} (search) vs. {4} ({1})'.format(search, label, std, len(search), len(std)))
         if match is not None:
             self.assertTrue(re.match(match, std, flags=re.MULTILINE) is not None,
-                msg + ' - failed to match {0}\n'.format(match) + label + ':' + std + '\nmatch expression: ' + match)
+                            msg + u' - failed to match in {1}:\n{2}\nmatch expression:\n{0}\nlengths: {3} (match) vs. {4} ({1})'.format(match, label, std, len(match), len(std)))
 
     def notok(self, response, msg='', search=None, match=None):
         if response.exitcode == 0:
@@ -887,3 +865,131 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         readcmdchunks = readcmd + ['--chunkhours=6']
         r = rpc(*readcmdchunks)
         self.ok(r, "read a lot of data with multiple reads")
+
+        r = rpc('flush', cik, rid1)
+        r = rpc('record', cik, rid1, '--value=41,12.33', '--value=42,12.34', '--value=43,12.35', '--value=-1,12.36')
+        self.ok(r, 'record some points')
+        r = rpc('read', cik, rid1, '--start=42', '--end=42', '--timeformat=unix', '--limit=3')
+        self.ok(r, 'read window of 1 second', match='42,12.34')
+
+        # TODO: why doesn't this match?
+        #r = rpc('read', cik, rid1, '--start=42', '--end=43', '--timeformat=unix', '--limit=3')
+        #self.ok(r, 'read window of 2 seconds', match='43,12.35\n42,12.34')
+
+        r = rpc('read', cik, rid1, '--start=44', '--timeformat=unix', '--limit=2')
+        self.ok(r, '--end has a default', match='[0-9]+,12.36')
+
+    def sort_test(self):
+        '''Read command with --sort flag'''
+        cik = self.client.cik()
+        rid = self._createMultiple(
+            cik,
+            [Resource(cik,
+                      'dataport',
+                      {'format': 'float', 'name': 'float_port'})])[0]
+
+        r = rpc('write', cik, rid, '--value=1')
+        self.ok(r, 'write 1')
+        time.sleep(1)
+        r = rpc('write', cik, rid, '--value=2')
+        self.ok(r, 'write 2')
+        time.sleep(1)
+        r = rpc('write', cik, rid, '--value=3')
+        self.ok(r, 'write 3')
+        time.sleep(1)
+
+        r = rpc('read', cik, rid, '--limit=2', '--format=raw', '--sort=asc')
+        self.ok(r, 'read two points, ascending', match="1\n2")
+
+        r = rpc('read', cik, rid, '--limit=2', '--format=raw', '--sort=desc')
+        self.ok(r, 'read two points, descending', match="3\n2")
+
+        r = rpc('read', cik, rid, '--limit=2', '--format=raw')
+        self.ok(r, 'read two points, ascending (desc is default)', match="3\n2")
+
+        r = rpc('read', cik, rid, '--limit=3', '--format=raw', '--sort=desc')
+        self.ok(r, 'read two points, descending', match="3\n2\n1")
+
+    def ip_test(self):
+        '''ip command'''
+        r = rpc('ip')
+        self.ok(r, 'ip command succeeds', match='\d+,\d+,\d+,\d+,\d+,\d+')
+
+    def data_test(self):
+        '''Data command'''
+        cik = self.client.cik()
+        rid_float, rid_str, rid_integer = self._createMultiple(
+            cik,
+            [Resource(cik, 'dataport',
+                      {'format': 'float'},
+                      alias='float'),
+             Resource(cik, 'dataport',
+                      {'format': 'string'},
+                      alias='string'),
+             Resource(cik, 'dataport',
+                      {'format': 'integer'},
+                      alias='integer')])
+
+        r = rpc('data', cik, '--read=float')
+        self.ok(r, 'read dataport with no value', '')
+
+        r = rpc('data', cik + 'a', '--write=float,3.1415')
+        self.notok(r, 'bad cik fails')
+
+        r = rpc('data', cik, '--write=float')
+        self.notok(r, 'bad write format fails')
+
+        r = rpc('data', cik, '--write=float,,2')
+        self.notok(r, 'multiple commas in write fails')
+
+        r = rpc('data', cik, '--write=float,3.1415')
+        self.ok(r, 'write single value', match='')
+
+        r = rpc('data', cik, '--read=float', '--read=integer')
+        self.ok(r, 'read two values, only one of which exists', match='float=3.1415')
+
+        r = rpc('data', cik, '--write=string,foo', '--write=integer,616')
+        self.ok(r, 'write multiple values', match='')
+
+        r = rpc('data', cik, '--read=string', '--read=integer', '--read=float')
+        self.ok(r, 'read multiple values', match='float=3.1415&integer=616&string=foo')
+
+        r = rpc('data', cik, '--read=integer', '--read=float', '--read=string')
+        self.ok(r, 'read multiple values, different order', match='string=foo&float=3.1415&integer=616')
+
+        r = rpc('data', cik, '--read=integer', '--write=float,17.5', '--write=string,bar', '--read=string')
+        # TODO: why does the platform return string=foo&integer=616 sometimes?
+        #self.ok(r, 'write and read multiple values', match='string=bar&integer=616')
+        self.ok(r, 'write and read multiple values', search='string=[a-z]{3}&integer=616')
+
+    def update_test(self):
+        '''Update command'''
+        cik = self.client.cik()
+
+        rid = self._createMultiple(
+            cik,
+            [Resource(cik, 'dataport',
+                      {'format': 'float', 'name': 'Original Name'},
+                      alias='float')])[0]
+
+        r = rpc('info', cik, rid, '--include=description')
+        info_old = json.loads(r.stdout)
+        name_old = info_old['description']['name']
+
+        r = rpc('update', cik, rid, '-', stdin=json.dumps({'name': 'update_test'}))
+        self.ok(r, 'update name', match='ok')
+
+        r = rpc('info', cik, rid, '--include=description')
+        info_new = json.loads(r.stdout)
+        name_new = info_new['description']['name']
+
+        log.debug(name_old)
+        log.debug(name_new)
+        log.debug('Description:')
+        log.debug(json.dumps(info_old))
+        log.debug(json.dumps(info_new))
+        self.assertTrue(name_old != name_new, 'client name was changed')
+        self.assertTrue(name_new == "update_test", 'new name is correct')
+
+        info_new['description']['name'] = name_old
+        self.assertTrue(info_old == info_new, 'client name was only difference')
